      program fortran_sqlite_demo
      implicit none

c     Shows how SQLite routines can be called from Fortran.
c     http://danial.org/sqlite/fortran/      A. Danial   July 2005
c     
c     Creates the database file /tmp/a.db with one table:
c        create table IXY(i    integer primary key, 
c                         x    float  ,
c                         y    float  ,
c                         city text   );
c     then inserts a bunch of rows into the table.  After populating
c     the database, performs a query on it and puts the results into 
c     Fortran variables.

      integer    MAX_COLS     ,  DATA_BUFFER
      parameter (MAX_COLS = 20,  DATA_BUFFER = 120000)
      integer    INT_TYPE  , REAL4_TYPE  , REAL8_TYPE  , STR_TYPE  ,
     .           INT_SIZE  , REAL4_SIZE  , REAL8_SIZE  , STR_SIZE
      parameter (INT_TYPE=1, REAL4_TYPE=2, REAL8_TYPE=3, STR_TYPE=4,
     .           INT_SIZE=1, REAL4_SIZE=1, REAL8_SIZE=2, STR_SIZE=10)

c     Strings and their integer overlays.
      integer          filename_overlay(20)
      character*80     filename
      equivalence     (filename, filename_overlay(1))
      integer          table_overlay(20)
      character*80     table
      equivalence     (table, table_overlay(1))
      integer          command_overlay(25)
      character*100    command
      equivalence     (command, command_overlay(1))

c     Data overlays.  All four arrays (cdata, idata, fdata, ddata)
c     will occupy the same region of memory.
      character*4      cdata(DATA_BUFFER)
      integer          idata(DATA_BUFFER)
      real*4           fdata(DATA_BUFFER)
      real*8           ddata(DATA_BUFFER/2)
      equivalence     (cdata(1), idata(1))
      equivalence     (cdata(1), fdata(1))
      equivalence     (cdata(1), ddata(1))

      character*40     cities(8)  ! arbitrary text data 

      integer          dbhandle,  ! need integer*8 for 64 bit executables
     .                 i, j, col, return_value, nRows, nCols, row_words,
     .                 offset, city_index, 
     .                 data_type(MAX_COLS), nWords(MAX_COLS),
     .                 offset4(MAX_COLS), offset8(MAX_COLS)

c     Routines in the the C wrapper file.
      integer          insert_data, open_database, close_database,
     .                 sql_begin, sql_commit, sql_insert, sql_do,
     .                 sql_query
      external         insert_data, open_database, close_database,
     .                 sql_begin, sql_commit, sql_insert, sql_do,
     .                 sql_query

      if (iargc() .eq. 1) then
         call getarg(1, filename)
      else
         write(6,*) 'Usage:  ./demo  <db filename>'
         stop
      end if
      call null_terminate(filename)

      command   = 'rm -f ' // filename
      table     = 'IXYT\0'
      cities(1) = 'Los Angeles                             '
      cities(2) = 'Rome                                    '
      cities(3) = 'Berlin                                  '
      cities(4) = 'London                                  '
      cities(5) = 'Madrid                                  '
      cities(6) = 'Quito                                   '
      cities(7) = 'Tokyo                                   '
      cities(8) = 'St. Petersburg                          '
      do i = 1,8
        call null_terminate(cities(i))
      end do

      write(6,*)  command
      call system(command) ! delete the database before trying to make it

c     Step 1:  Create a database.
      return_value = open_database(filename_overlay, dbhandle)

      command      = 'create table IXYT(i integer primary key, x float, 
     . y float, city text);\0'
      return_value = sql_do(dbhandle, command_overlay)

c     Step 2:  Populate the database.
      nCols = 4
      data_type(1) =  INT_TYPE   ! i is integer
      data_type(2) =  REAL4_TYPE ! x is real*4
      data_type(3) =  REAL8_TYPE ! y is real*8
      data_type(4) =  STR_TYPE   ! t is a string (40 chars = 10 words)
      nWords(1)    =  INT_SIZE
      nWords(2)    =  REAL4_SIZE
      nWords(3)    =  REAL8_SIZE
      nWords(4)    =  STR_SIZE

c     Assign values to offset arrays to make subsequent indexing
c     a bit easier.
      row_words  = nWords(1)
      offset4(1) = 1
      offset8(1) = 1
      do i = 2,nCols
        row_words  =  row_words    + nWords(i) 
        offset4(i) =  offset4(i-1) + nWords(i-1) 
        offset8(i) = (offset4(i) + 1)/2
      end do
      nRows = DATA_BUFFER/(row_words + 1)

c     Fill the transfer buffer with data.  This is the most difficult
c     and error-prone part of the task.  Misaligned data will cause all
c     kinds of grief.
c       i       will be 1..N
c       x and y will be   random values on -50.0 .. +50.0
c       city    will be a random entry out of the cities() array
      do i = 0,nRows-1
        idata(offset4(1) + row_words*i  ) = i + 1               ! i
        fdata(offset4(2) + row_words*i  ) = 50.0 - 100.0*rand() ! x
        ddata(offset8(3) + row_words*i/2) = 50.0 - 100.0*rand() ! y
        city_index = 1 + int(8.0*rand())  ! pick a random city
        do j = 0,9
          cdata(offset4(4) + row_words*i+j) = 
     .      cities(city_index)(4*j+1:4*j+4)
        end do
      end do

      return_value = sql_insert(dbhandle     , 
     .                          table_overlay,
     .                          nRows        ,
     .                          nCols        ,
     .                          data_type    ,
     .                          nWords       ,
     .                          idata        )
      write(6,*) 'Created database ' // filename

c     Step 3:  Query the database and grab the results into Fortran 
c     variables.  First empty the buffer to make the query more 
c     convincing.
      do i = 1,DATA_BUFFER
        idata(i) = 0  ! which also nulls out fdata, ddata, cdata
      end do
      nRows = 0

      command = 'select avg(y),count(i),city from ixyt group by city ord
     .er by city;\0'
      nCols   = 3  ! the result will have three columns
      data_type(1) = REAL4_TYPE  ! bring y back as a real*4 (could just
      data_type(2) = INT_TYPE    ! as easily do real*8)
      data_type(3) = STR_TYPE
      return_value = sql_query(dbhandle        , ! in
     .                         command_overlay , ! in
     .                         nCols           , ! in
     .                         data_type       , ! in
     .                         nRows           , ! out
     .                         idata)            ! out
      write(6,*) command,' returned ', nRows, ' rows of data'
      offset = 1
      do i = 1,nRows
        write(6,'(a,i3,a,$)') 'row ', i, ':   '
        do col = 1,nCols
          if      (data_type(col) .eq. 1) then  
            write(6,'(i8,a,$)') idata(offset), ' '
            offset = offset + 1
          else if (data_type(col) .eq. 2) then  
            write(6,'(1pe15.8,a,$)') fdata(offset), ' '
            offset = offset + 1
          else if (data_type(col) .eq. 3) then  
            write(6,'(1pe22.16,a,$)') ddata(offset/2), ' '
            offset = offset + 2
          else if (data_type(col) .eq. 4) then  
            do j = 0,9
              write(6,'(a,$)') cdata(offset)(4*j+1:4*j+4)
            end do
            write(6,'(a,$)') ' '
            offset = offset + 10
          else
            write(6,*)
            write(6,*) 'Unknown data type, ', data_type(col),
     .                 ' for column ',col
          end if
        end do
        write(6,*)
      end do

      return_value = close_database(dbhandle)

      end

      subroutine null_terminate(string)
      character*(*) string
      integer i

      do i = len(string), 1, -1
        if (string(i:i) .ne. ' ') then
          goto 10 ! break
        else
          string(i:i) = '\0'
        end if
      end do
   10 continue
      return
      end
